# Built-in
import unittest
import os

# Internal
from nxt.plugins.file_fallbacks import NXT_FILE_ROOTS
from nxt.session import Session
from nxt.stage import DATA_STATE


class TestFallbacks(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

    def test_fallbacks(self):
        print("Test that fallbacks works.")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        cwd = "../test/plugins/fallbacks/base"
        expanded = os.path.abspath(cwd)
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init/node')
        expected = os.path.join(expanded, 'biped_base.txt').replace(os.sep, '/')
        attr_name = 'filepath'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)

    def test_multi_file_tokens(self):
        print("Test that an attr with multiple file tokens resolves as "
              "expected (using Stage.resolve_file_token).")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        p1 = os.path.abspath(os.path.dirname(graph)).replace(os.sep, '/')
        p2 = os.path.abspath(os.path.join(p1, '..', 'base',
                                          'biped_base.txt')).replace(os.sep,
                                                                     '/')
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init/node2')
        expected = "['{}', '', '{}']".format(p1, p2)
        attr_name = 'limitation'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)

    def test_instance_fallbacks(self):
        print("Test that fallback work with instances.")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        fb_graph = "../test/plugins/fallbacks/base/fallbacks2.nxt"
        expanded = os.path.abspath(os.path.dirname(fb_graph))
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init2/node')
        expected = os.path.join(expanded, 'biped_base.txt').replace(os.sep, '/')
        attr_name = 'filepath'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)

    def test_code_file_tokens(self):
        print("Test that fallback give full path in code")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        expanded = os.path.abspath(os.path.dirname(graph))
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init')
        expected = os.path.join(expanded, 'dummy.txt').replace(os.sep, '/')
        actual = self.stage.get_node_code_string(node, comp_layer,
                                                 DATA_STATE.RESOLVED)
        self.assertEqual(expected, actual)

    def test_muted_layers_fallbacks(self):
        print("Test that fallback respects muted layers.")
        # Test base without muting anything
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        expanded = os.path.abspath(os.path.dirname(graph))
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/test_mute')
        expected = os.path.join(expanded, 'dummy.txt').replace(os.sep, '/')
        attr_name = 'filepath'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)
        # Mute and test again
        self.stage.top_layer.set_muted(True)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/test_mute')
        expanded2 = os.path.abspath("../test/plugins/fallbacks/base")
        expected = os.path.join(expanded2, 'biped_base.txt').replace(os.sep,
                                                                     '/')
        attr_name = 'filepath'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)

    def test_file_list(self):
        print("Test file list token works.")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        cwd = "../test/plugins/fallbacks/base"
        expanded = os.path.abspath(cwd)
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init/node')
        expected = [os.path.join(expanded,
                                 'biped_base.txt').replace(os.sep, '/'),
                    os.path.join(expanded,
                                 'quad_base.txt').replace(os.sep, '/')
                    ]
        attr_name = 'file_list'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(str(expected), actual)

    def test_nested_file_tokens(self):
        print("Test nested file tokens")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        cwd = "../test/plugins/fallbacks/another"
        expanded = os.path.abspath(cwd)
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init/node')
        expected = os.path.join(expanded, 'dummy.txt').replace(os.sep, '/')
        attr_name = 'nested'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)

    def test_env_var(self):
        print("Test that NXT_FILE_ROOTS env var works.")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        cwd = "../test/plugins/fallbacks/base"
        expanded = os.path.abspath(cwd)
        os.environ['NXT_FILE_ROOTS'] = expanded
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init/node3')
        expected = os.path.join(expanded, 'biped_base.txt').replace(os.sep, '/')
        attr_name = 'env'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        os.environ.pop('NXT_FILE_ROOTS')
        self.assertEqual(expected, actual)

    def test_token_resolve_to_nothing(self):
        print("Test a token containing failed attr subs returns nothing.")
        graph = "../test/plugins/fallbacks/another/fallbacks_top.nxt"
        cwd = "../test/plugins/fallbacks/base"
        expanded = os.path.abspath(cwd)
        self.stage = Session().load_file(graph)
        comp_layer = self.stage.build_stage()
        node = comp_layer.lookup('/init/node4')
        expected = ''
        attr_name = 'nothing'
        actual = self.stage.resolve(node, getattr(node, attr_name),
                                    comp_layer, attr_name=attr_name)
        self.assertEqual(expected, actual)


class TestPathWithRoots(unittest.TestCase):
    def setUp(self):
        self.session = Session()
        self.test_filename = 'NotAFile.txt'
        self.test_input = '${path::' + self.test_filename + '}'

    def test_empty(self):
        print('Test that in an empty file with no environment,'
              ' path:: resolves to cwd')
        test_stage = self.session.new_file()
        test_node, _ = test_stage.add_node()
        test_node = test_node[0]
        test_attr = test_stage.add_node_attr(test_node, 'path_test',
                                             {'value': self.test_input},
                                             test_stage.top_layer)
        my_dir = os.path.dirname(os.path.realpath(__file__))
        old_cwd = os.getcwd()
        # Keep any environment out of this test
        old_env_roots = ''
        if NXT_FILE_ROOTS in os.environ:
            old_env_roots = os.environ.pop(NXT_FILE_ROOTS)
        os.chdir(my_dir)
        expected = os.path.join(my_dir, self.test_filename).replace(os.sep, '/')
        found = test_stage.get_node_attr_value(test_node, test_attr,
                                               test_stage.top_layer,
                                               resolved=True)
        self.assertEqual(expected, found)
        os.chdir(old_cwd)
        os.environ[NXT_FILE_ROOTS] = old_env_roots

    def test_saved(self):
        print('Test in a saved file with no environment, path:: '
              'resolves to graph dir')
        # Keep any existing environment out of this test
        old_env_roots = ''
        if NXT_FILE_ROOTS in os.environ:
            old_env_roots = os.environ.pop(NXT_FILE_ROOTS)
        old_cwd = os.getcwd()
        my_dir = os.path.dirname(os.path.realpath(__file__))
        os.chdir(my_dir)

        graph_path = "../test/plugins/fallbacks/pathroots.nxt"
        graph_path = os.path.abspath(graph_path)
        graph_dir = os.path.dirname(graph_path)
        test_stage = self.session.load_file(graph_path)
        test_node = test_stage.top_layer.lookup('/test_node')
        test_attr = test_stage.add_node_attr(test_node, 'path_test',
                                             {'value': self.test_input},
                                             test_stage.top_layer)
        expected = os.path.join(graph_dir, self.test_filename).replace(os.sep,
                                                                       '/')
        found = test_stage.get_node_attr_value(test_node, test_attr,
                                               test_stage.top_layer,
                                               resolved=True)
        self.assertEqual(expected, found)

        print('Test in a saved file with an environment, path:: '
              'resolves to first real env root')
        test_root = os.path.dirname(my_dir)
        os.environ[NXT_FILE_ROOTS] = 'a super fake root;' + test_root
        expected = os.path.join(test_root, self.test_filename).replace(os.sep,
                                                                       '/')
        found = test_stage.get_node_attr_value(test_node, test_attr,
                                               test_stage.top_layer,
                                               resolved=True)
        self.assertEqual(expected, found)

        os.chdir(old_cwd)
        os.environ[NXT_FILE_ROOTS] = old_env_roots




# Builin
import os
import sys

import subprocess
import logging.handlers
import json
import socket

if sys.version_info[0] == 2:
    from SocketServer import ThreadingMixIn
    from SimpleXMLRPCServer import SimpleXMLRPCServer
else:
    from socketserver import ThreadingMixIn
    from xmlrpc.server import SimpleXMLRPCServer

# Internal
import nxt
from nxt.stage import GraphError
from nxt.remote import get_server_info_filepath
import nxt.remote.contexts
from nxt import nxt_path, nxt_io, nxt_log

server = None
logger = logging.getLogger('nxt')


class NxtServerException(GraphError):
    pass


class NxtServer(ThreadingMixIn, SimpleXMLRPCServer):
    def __init__(self, address, allow_none=True, log_requests=False):
        """Simple Threaded XMLRPCServer.
        :param address: tuple of (HOST, PORT)
        :param allow_none: If True None type objects are marshaled.
        :param log_requests: If True connection requests are logged.
        """
        SimpleXMLRPCServer.__init__(self, addr=address, allow_none=allow_none,
                                    logRequests=log_requests)


class ServerFunctions(object):
    def __init__(self, log_filepath):
        self.cache_file = None
        self.log_filepath = log_filepath

    def get_log_location(self):
        """Simple method for getting the filepath of the file the server is
        logging to. May be an empty string if the server isn't logging to a
        custom filepath.
        :return: str
        """
        return self.log_filepath or ''

    @staticmethod
    def is_alive():
        """Simple method for clients to use to check if the server is still
        alive.
        :return: True
        """
        logger.info("rpc server is alive and ready")
        return True

    @staticmethod
    def exec_in_headless(filepath, start_node, cache_path,
                         parameters, context_name):
        """Executed the given graph (filepath) with the given start_node in the
        dcc exe (as a sub-process). The temp path (if provided) must be a
        location that the server can read/write. If no temp path is given one
        will be generated. The file at the temp path is used to store the
        cache data to be returned to the caller. Only the file path is
        returned to the caller, not the actual data.
        :param filepath: Path to nxt save file
        :param start_node: start node path
        :param cache_path: Path to store output cache data (if none is given
        one will be generated)
        :param parameters: Optional parameters dict
        :param context_name: name of context, defaults to python
        :return: filepath to temp file
        """
        # Fixme: Contexts must be accessed like this to avoid importing them
        #  again and thus emptying the list of user contexts.
        context = nxt.remote.contexts.find_context_by_name(context_name)
        if not context:
            known = nxt.remote.contexts.iter_context_names()
            logger.debug('Known contexts: \n{}'.format('\n'.join(known)))
            raise NameError('Unknown context "{}"'.format(context_name))
        context_exe = context.exe
        if not context or not context_exe:
            raise TypeError('Unable to find context exe for: '
                            '{}'.format(context))

        context_graph = context.graph
        context_graph = nxt_path.full_file_expand(context_graph)
        if not context_graph:
            raise TypeError('No launch script found for context: '
                            '{}'.format(context))
        # Setup cache file if none provided
        if not cache_path:
            cache_path = nxt_io.generate_temp_file()
        cache_path = cache_path.replace(os.sep, '/')
        # Setup parameters file if parameters provided
        if parameters:
            parameters_file = nxt_io.generate_temp_file()
            parameters_file = parameters_file.replace(os.sep, '/')
            with open(parameters_file, 'w+') as fp:
                json.dump(parameters, fp, separators=(',', ': '))
        else:
            parameters_file = ''
        context_graph = context_graph.replace(os.sep, '/')
        logger.info('Starting \n'
                    'Context: {} \n'
                    'Interpreter: {} \n'
                    'Context Graph: {}\n'.format(context, context_exe,
                                                 context_graph))
        logger.info('Cache location: {}\n'.format(cache_path))
        # Format the cli call
        safe_graph_path = nxt_path.full_file_expand(filepath)
        # open context with graph and parameters
        os.environ[nxt_log.VERBOSE_ENV_VAR] = 'socket'
        cli_args = ['exec', context_graph, '-p',
                    '/.graph_file', safe_graph_path,
                    '/.cache_file', cache_path]
        if parameters_file:
            cli_args += ['/.parameters_file', parameters_file]
        script = os.path.join(os.path.dirname(__file__), '..', 'cli.py')
        script = os.path.abspath(script)
        script = script.replace(os.sep, '/')
        if context.args:
            extra_args = []
            if context.args:
                extra_args = list(context.args)
            args = [context_exe] + extra_args + [script, '--'] + cli_args
        else:
            args = [context_exe, script] + cli_args
            if start_node:
                args += ['/enter/call_graph._start', start_node]
        # HACK solution only until refined context system rolls out to relate
        # format strings to context names to include space for cli args.
        if 'UE4Editor' in context_exe:
            args = context_exe.format(cli_path=script,
                                      cli_args=' '.join(cli_args))

        logger.debug('call:  {}'.format(args))
        # TODO: Find a clean way to raise exceptions from the subprocess.
        try:
            subprocess.call(args)
        except subprocess.CalledProcessError:
            raise RuntimeError('Remote context graph "{}" failed! See log...'
                               ''.format(safe_graph_path))
        return cache_path

    def kill(self):
        """Shuts down the running rpc server and attemps to remove the server
        log file if there is one.
        :return: None
        """
        global server
        if server:
            logger.warning('Shutting down rpc server!')
            server.shutdown()
            if self.log_filepath:
                try:
                    logger.debug('Removing server log file: '
                                 '{}'.format(self.log_filepath))
                    os.remove(self.log_filepath)
                except:
                    logger.warning('Failed to remove server log file, '
                                   'if it was an autogenerated file nxt_log '
                                   'will remove it when determined it is '
                                   'stale.')


def secure_address():
    server_info_filepath = get_server_info_filepath()
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    _, port = s.getsockname()
    with open(server_info_filepath, 'w+') as fp:
        fp.write('localhost' + '\n')
        fp.write(str(port))
    return 'localhost', port


def run_server(address=None, log_requests=False):
    nxt_root = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '../..')).replace(os.sep, '/')
    os.chdir(nxt_root)
    global server
    if address is None:
        address = secure_address()
    server = NxtServer(address, log_requests=log_requests)
    server.register_instance(ServerFunctions(log_filepath=None))
    server.allow_reuse_address = True
    logger.info('Threaded nxt rpc server started!')
    logger.debug('Listening on:  {}'.format(address))
    server.serve_forever()


if __name__ == '__main__':
    logger.info('Available contexts: '
                '{}'.format(list(nxt.remote.contexts.iter_context_names())))
    # TODO: Fix logging so non-hosting apps can get the stdout and logs
    # logger.debug('Logging to: {}'.format(log_file))
    logger.info('Starting up server...')
    run_server()
    # logger.info('Logging rpc stdout to: {}'.format(log_file))



# Builtin
import logging
import os
import unittest
import time

# Internal
from nxt.session import Session
from nxt import nxt_path, nxt_io
from nxt.tokens import (register_token, get_token_content,
                        get_standalone_tokens, make_token_str, TOKENTYPE)
from nxt.nxt_layer import CompLayer, get_active_layers
from nxt.nxt_node import get_node_path


logger = logging.getLogger('nxt')

FILEPATH_PREFIX = TOKENTYPE.FILE.prefix
FILE_LIST_PREFIX = 'filelist::'

NXT_FILE_ROOTS = 'NXT_FILE_ROOTS'
ROOT_PATH_SEP = os.pathsep

INVALID_PATH = '<invalid>'


def detect_filepath_token(value):
    return value.startswith(FILEPATH_PREFIX)


def detect_file_list_token(value):
    return value.startswith(FILE_LIST_PREFIX)


def resolve_filepath_token(stage, node, cleaned, layer, **kwargs):
    """Custom resolver for `file::` tokens. Can resolve multiple file tokens
    in a single attr value. The initial raw value is NOT retrieved from the
    node again as fallbacks happen.
        If on layer_0 the attr value is
            ${file::${attr}}
        and on layer_1 the attr value is
            ${file::${foo}}
        This plugin will only fallback and resolve for the value:
              ${file::${attr}}
    This is a fallback resolver for the contents of the token, not the entire
    attr value!
    :param stage: Stage instance
    :param node: Node object
    :param cleaned: String of cleaned attr value (the outer file
    token has been stripped off).
    :param layer: Layer object
    :param kwargs: Not used but may be passed by the Stage
    :return: string of resolved file path
    """
    token = make_token_str(FILEPATH_PREFIX+cleaned)
    out_value = resolve_file_token_with_fallbacks(stage=stage,
                                                  comp_node=node,
                                                  value=token,
                                                  comp_layer=layer)
    return out_value


def resolve_filelist_token(stage, node, cleaned, layer, **kwargs):
    as_string = kwargs.get('as_string', True)
    token = make_token_str(FILE_LIST_PREFIX+cleaned)
    genny = resolve_file_list_token(stage=stage, comp_node=node,
                                    value=token, comp_layer=layer)
    file_list = list(genny)
    if as_string:
        return str(file_list)
    return file_list


def resolve_file_list_token(stage, comp_node, value, comp_layer):
    if not isinstance(comp_layer, CompLayer):
        logger.error('File list resolve requires a comp layer or a '
                     'sub-class of a comp layer!')
        return stage.resolve_file_token(comp_node, value, comp_layer)
    first, last = comp_layer._layer_range
    paths = generate_historical_values(first, last, value, stage,
                                       comp_node, comp_layer, file_list_token)
    return paths


def resolve_file_token_with_fallbacks(stage, comp_node, value, comp_layer):
    """Given a node, who's attr_name has a value only containing a file
    token we will fall back through all possible historical comps until
    we bottom out or find a valid file path.
    :param stage: Stage instance to resolve against
    :param comp_node: CompNode object
    :param value: string of attribute value
    :param comp_layer: CompLayer object
    :raises: TypeError
    :return: string of file path or empty string if no file found
    """
    start = time.time()
    if not isinstance(comp_layer, CompLayer):
        raise TypeError('Resolve with fallbacks requires a comp layer or a '
                        'sub-class of a comp layer!')
    first, last = comp_layer._layer_range
    paths = generate_historical_values(first, last, value, stage,
                                       comp_node, comp_layer, file_fallback_token)
    try:
        valid_path = next(paths)
    except StopIteration:
        valid_path = ''
    # resolve_time = str(int(round((time.time() - start) * 1000)))  # debugging
    return valid_path


def generate_historical_values(first, last, value, stage, comp_node,
                               comp_layer, token):
    seen = []
    for depth in range(first, last):
        requested_layers = stage._sub_layers[depth:last]
        # Early exit if there are no active layers in the requested range
        active_layers = get_active_layers(requested_layers)
        if not active_layers:
            continue
        historical_node = stage.infer_lower_comp(comp_node, comp_layer,
                                                 active_layers=active_layers,
                                                 depth=depth)
        resolved = historical_resolve_file_ref(stage, historical_node,
                                               value, comp_layer,
                                               active_layers,
                                               historical_depth=depth,
                                               token=token)
        if resolved and resolved not in seen:
            yield resolved
            seen += [resolved]


def historical_resolve_file_ref(stage, comp_node, value, comp_layer,
                                active_layers, token, historical_depth=0):
    """Attempts to resolve a file token in the attribute
    specified at the historical depth given. See `Stage.infer_lower_comp` for
    more details on comping at a historical depth. This method is not
    recursive, it only checks the exact historical depth provided.
    :param stage: Stage instance to resolve against
    :param comp_node: CompNode object
    :param value: string of attribute value
    :param comp_layer: CompLayer object
    :param active_layers: List of at least 1 active layer
    :param token: tokens.Token object
    :param historical_depth: number of layers to step back, relative to the
    comp layer, that the historical comp should be built at.
    :return: string of the resolved attribute value
    """
    resolved = value
    display_layer = active_layers[0]
    layer_dir = display_layer.get_cwd()
    all_tokens = get_standalone_tokens(value, token_types=(token,))
    # If no tokens are found we test if the val is a valid a valid path
    if not all_tokens and value:
        resolved = file_root_fallback(resolved, layer_dir=layer_dir)
    for full_token in all_tokens:
        token_content = get_token_content(full_token)
        string = token_content
        _, sep, raw = string.partition(token.prefix)
        resolved_token = historical_resolve_attr_ref(stage, raw, comp_node,
                                                     comp_layer, active_layers,
                                                     historical_depth)
        if not resolved_token and raw:
            resolved_token = INVALID_PATH
        resolved_fp = file_root_fallback(resolved_token, layer_dir=layer_dir)
        resolved = resolved.replace(full_token, resolved_fp)
    return resolved


def historical_resolve_attr_ref(stage, string, comp_node, comp_layer,
                                active_layers, historical_depth=0):
    """Attempts to resolve an attr ref token at the historical depth given.
    See `Stage.infer_lower_comp` for more details on comping at a historical
    depth. This method IS recursive, BUT only checks the exact historical
    depth provided.
    :param stage: Stage instance
    :param string: Attr value string containing 1 or more attr ref tokens.
    :param comp_node: CompNode object
    :param comp_layer: CompLayer object
    :param active_layers: list of active layers
    :param historical_depth: number of layers to step back, relative to the
    comp layer, that the historical comp should be built at.
    :return: string of the resolved attribute token
    """
    tokens = stage.get_tokens_from(string, token_type=TOKENTYPE.ATTR)
    resolved_token = string
    for attr_token in tokens:
        attr_ref = get_token_content(attr_token)
        ref_path, ref_attr_name = nxt_path.path_attr_partition(attr_ref)
        if ref_path and ref_attr_name:
            attr_name = ref_attr_name
            source_path = ref_path
        else:
            attr_name = attr_ref
            source_path = get_node_path(comp_node)
        # If this attr ref is to another node, go there and get the value
        if source_path:
            node_path = get_node_path(comp_node)
            source_path = nxt_path.expand_relative_node_path(source_path,
                                                             node_path)
        source_node = comp_layer.lookup(source_path)
        historical_node = stage.infer_lower_comp(source_node, comp_layer,
                                                 active_layers,
                                                 historical_depth)

        value = getattr(historical_node, attr_name, None) or ''
        more = get_standalone_tokens(value, token_types=(TOKENTYPE.ATTR,
                                                         file_fallback_token))
        for token_content in more:
            token = stage.determine_token_type(get_token_content(token_content))
            if token == TOKENTYPE.ATTR:
                value = historical_resolve_attr_ref(stage, value, comp_node,
                                                    comp_layer,
                                                    active_layers,
                                                    historical_depth)
            elif token == file_fallback_token:
                value = historical_resolve_file_ref(stage, comp_node,
                                                    token_content, comp_layer,
                                                    active_layers,
                                                    token,
                                                    historical_depth)
        raw_token = make_token_str(attr_ref)
        resolved_token = resolved_token.replace(raw_token, value)
    return resolved_token


def iter_env_roots(only_exsiting=True):
    if NXT_FILE_ROOTS not in os.environ:
        return
    if NXT_FILE_ROOTS in os.environ:
        for root in os.environ.get(NXT_FILE_ROOTS).split(ROOT_PATH_SEP):
            if only_exsiting and not os.path.exists(root):
                continue
            yield root


def file_root_fallback(filepath, layer_dir=None):
    """Test a given filepath against the list of possible root paths
    specified in the env var NXT_FILE_ROOTS.
    :param filepath: string of file path
    :param layer_dir: optional layer dir to be tested after all possible
    roots fail.
    :return: string of valid filepath or empty string
    """
    for root in iter_env_roots():
        full_path = nxt_path.full_file_expand(filepath, start=root)
        if os.path.exists(full_path):
            return full_path
    if layer_dir and os.path.exists(layer_dir):
        full_path = nxt_path.full_file_expand(filepath, start=layer_dir)
        if os.path.exists(full_path):
            return full_path
    return ''


nxt_io.register_reference_path_expander(file_root_fallback)


file_fallback_token = register_token(FILEPATH_PREFIX, detect_filepath_token,
                                     resolve_filepath_token)

file_list_token = register_token(FILE_LIST_PREFIX, detect_file_list_token,
                                 resolve_filelist_token)


def detect_path_token(value):
    return value.startswith(TOKENTYPE.FILEPATH.prefix)


def resolve_path_token_in_roots(stage, node, cleaned, layer, **kwargs):
    value = stage.resolve(node, cleaned, layer)
    try:
        root = next(iter_env_roots())
    except StopIteration:
        layer_cwd = layer.get_cwd()
        if os.path.exists(layer_cwd):
            root = layer_cwd
        else:
            root = os.getcwd()
    return nxt_path.full_file_expand(value, start=root)


roots_path_token = register_token(TOKENTYPE.FILEPATH.prefix, detect_path_token,
                                  resolve_path_token_in_roots)


# UNITTESTS
# Fixme: putting these here until we can make plugins packages



"""Code used while a graph is running.
"""
# Built-in
import code
import sys
import traceback
import logging

# Internal
from . import IGNORE

logger = logging.getLogger(__name__)

def w(string, quote_type=0):
    """Wraps given string in quotes.

    :param string: Input string to be wrapped
    :type string: str
    :param quote_type: Int 1: ', 2: ", 3: ''', 4: \"\"\" Default is 1, any \
        string can also be provided and we will wrap the string arg in that \
        string ie w('Hello World', '$') returns '$Hello World$'.
    :type quote_type: int or str
    :return: String wrapped in quote marks or custom string
    """
    raise NotImplementedError("Written here for documentation, real "
                              "implementation in stage.py")

def execute(paths=(), start=None, parameters=None):
    """While within node code, this method runs other parts of the graph, \
        allowing non-linear execution.

    :param paths: iterable of paths to execute, defaults to ()
    :type paths: iterable, optional
    :param start: node path to start from, defaults to None. \
        NOTE must give *either* paths or start, both is invalid input.
    :type start: str, optional
    :param parameters: parameters dictionary to apply before exectuion, \
        defaults to None
    :type parameters: dict, optional
    :raises GraphError: When there is a problem with execution.
    :raises ValueError: When not given either node paths or a start path.
    """
    raise NotImplementedError("Written here for documentation, real "
                              "implementation in stage.py")

class Console(code.InteractiveConsole):
    def __init__(self, _globals=None, _locals=None, filename=IGNORE,
                 node_path=None, layer_path=None):
        if locals is None:
            _locals = {}
        if globals is None:
            _globals = {}
        self.node_path = node_path
        self.running_lines = []
        self.layer_path = layer_path
        self.locals = _locals
        self.globals = _globals
        self.run_as_global = False
        self.lineno_offset = -1
        code.InteractiveConsole.__init__(self, self.locals, filename)

    def runcode(self, c):
        try:  # Convert to tuple for python3
            exec(c, self.globals)
        except (KeyboardInterrupt, ExitNode, ExitGraph):
            raise
        except (SystemExit, Exception) as err:
            lineno = get_traceback_lineno(err_depth=1)
            lineno -= 1
            try:
                bad_line = self.running_lines[lineno-1]
            except IndexError:
                bad_line = "LINE NOT FOUND"
            _, _, tb = sys.exc_info()
            raise GraphError(err, tb, self.layer_path, self.node_path, lineno,
                             bad_line, err_depth=1)


def get_traceback_lineno(err_depth=0):
    """Get the line number of one of the errors in the current traceback.

    :param err_depth: index of the error to get line number of. 0 is most
    recent, higher values are deeper errors. Defaults to 0
    :type err_depth: int, optional
    :return: line number of requested error
    :rtype: int
    """
    _, _, tb = sys.exc_info()
    _tblist = traceback.extract_tb(tb)
    return _tblist[err_depth][1]


class GraphError(Exception):
    def __init__(self, err, tb, layer_path, err_path, lineno, bad_line,
                 err_depth=0):
        tb_lines = traceback.extract_tb(tb)
        tb_lines = tb_lines[err_depth:]
        # Insert our custom graph traceback
        tb_lines[0] = (layer_path, lineno, err_path, bad_line)
        print_lines = traceback.format_list(tb_lines)
        print_lines[0] = print_lines[0].lstrip()
        print_lines += traceback.format_exception_only(type(err), err)
        print_tb = ''.join(print_lines)
        print_tb = print_tb.rstrip('\n')
        super(Exception, self).__init__(print_tb)


class GraphSyntaxError(GraphError):
    def __init__(self, error, layer_path, error_path, lineno):
        print_lines = traceback.format_exception_only(type(error), error)
        top_line_fmt = 'File "{}", line {}, in {}\n'
        print_lines[0] = top_line_fmt.format(layer_path, lineno, error_path)
        syntax_err_msg = ''.join(print_lines)
        syntax_err_msg = syntax_err_msg.rstrip('\n')
        super(Exception, self).__init__(syntax_err_msg)


class InvalidNodeError(GraphError):
    def __init__(self, node_path):
        super(Exception, self).__init__("Attempted to execute "
                                        "non-exsistant node! \n"
                                        "{}".format(node_path))


class ExitNode(Exception):
    def __init__(self, reason=''):
        super(ExitNode, self).__init__(reason)


class ExitGraph(Exception):
    def __init__(self, reason=''):
        self.runtime_layer = None
        super(ExitGraph, self).__init__(reason)



# Built-in
import unittest
import os
import logging

# Internal
from nxt.nxt_layer import LayerReturnTypes
from nxt import DATA_STATE
from nxt.nxt_node import INTERNAL_ATTRS
from nxt.session import Session

root_logger = logging.getLogger('nxt')
root_logger.propagate = False
root_logger.addHandler(logging.NullHandler())


def create_spec_graph_test_func(filename):
    def spec_test(self):
        stage = Session().load_file(filepath=filename)
        r = LayerReturnTypes.Node
        for input_node in stage.top_layer.descendants(return_type=r):
            # Only concerned with test input nodes
            name = getattr(input_node, INTERNAL_ATTRS.NAME)
            if not name.startswith('test'):
                continue
            if not name.endswith('input'):
                continue
            input_path = stage.top_layer.get_node_path(input_node)
            rt_layer = stage.execute(input_path)
            # Gather expectations
            expected_resolve_path = input_path.replace('input', 'resolved')
            expected_resolve_node = stage.top_layer.lookup(expected_resolve_path)
            expected_cache_path = input_path.replace('input', 'cached')
            expected_cache_node = stage.top_layer.lookup(expected_cache_path)
            cached_node = rt_layer.lookup(input_path)
            # Assert computes are equal
            resolved_code = stage.get_node_code_string(input_node,
                                                             layer=stage.top_layer,
                                                             data_state=DATA_STATE.RESOLVED)
            expected_resolve_code = stage.get_node_code_string(expected_resolve_node,
                                                                     layer=stage.top_layer.lookup,
                                                                     data_state=DATA_STATE.RAW)
            self.assertEqual(expected_resolve_code, resolved_code)
            cached_code = getattr(cached_node, INTERNAL_ATTRS.CACHED_CODE)
            expected_cached_code = stage.get_node_code_string(expected_cache_node,
                                                                    layer=rt_layer,
                                                                    data_state=DATA_STATE.RAW)
            self.assertEqual(expected_cached_code, cached_code)
            # Assert attrs are equal
            for attr_name in stage.get_node_attr_names(input_node):
                resolved_attr_val = stage.get_node_attr_value(input_node,
                                                              attr_name,
                                                              stage.top_layer,
                                                              resolved=True)
                excpected_resolve_attr_val = stage.get_node_attr_value(expected_resolve_node,
                                                                       attr_name,
                                                                       stage.top_layer,
                                                                       resolved=False)
                self.assertEqual(excpected_resolve_attr_val, resolved_attr_val)
                cached_attr_val = getattr(cached_node, attr_name)
                expected_cached_attr_val = stage.get_node_attr_value(expected_cache_node,
                                                                     attr_name,
                                                                     stage.top_layer)
                self.assertEqual(expected_cached_attr_val, cached_attr_val)
    return spec_test


class SpecGraphTest(unittest.TestCase):
    pass


test_dir = os.path.dirname(__file__)
SPEC_GRAPHS_DIR = os.path.join(test_dir, 'spec_graphs')

for filename in os.listdir(SPEC_GRAPHS_DIR):
    test_name = 'test_{}'.format(os.path.basename(filename))
    path = os.path.join(SPEC_GRAPHS_DIR, filename)
    func = create_spec_graph_test_func(path)
    func.__name__ = test_name
    setattr(SpecGraphTest, test_name, func)



# Builtin
import unittest
import json
import os
import re

# Internal
from nxt import nxt_path
from nxt.constants import GRAPH_VERSION
from nxt.session import Session

os.chdir(os.path.dirname(os.path.abspath(__file__)))


class LegacyFileConversion(unittest.TestCase):

    def test_auto_update(self):
        self.maxDiff = None
        self.stage = Session().load_file("legacy/0.45.0_TopLayer.nxt")
        self.file_data = self.stage.get_layer_save_data(0)
        self.file_data = json.dumps(self.file_data, indent=4,
                                    sort_keys=False, separators=(',', ': '))
        file_path = "legacy/0.45.0_to_LATEST.nxt"
        real_path = nxt_path.full_file_expand(file_path)
        self.proof_data = dynamic_version(real_path)
        self.assertEqual(self.proof_data, self.file_data)


def dynamic_version(real_path, version=GRAPH_VERSION.VERSION_STR):
    version = '"' + version + '"'
    with open(real_path, 'r') as file_object:
        json_data = file_object.read()
        pattern = r"((\"version\": )([0-9\.\"]*)(\,))"
        ref, num = re.search(pattern, json_data).group(0, 3)
        cur_version = ref.replace(num, version)
        proof_data = json_data.replace(ref, cur_version)
        return proof_data



# Builtin
import os
import inspect

# NOTE: the behavior of inspect.currentframe() likely requires cpython.
# https://docs.python.org/2.7/library/inspect.html#inspect.currentframe
_this_file = inspect.getframeinfo(inspect.currentframe()).filename
TEST_DIR = os.path.dirname(os.path.abspath(_this_file))


def get_test_file_path(file_name):
    """Shortcut to get full path to desired test file.

    Only assembles path, no validation.

    :param file_name: Filename of testing file.
    :type file_name: str
    :return: Full path for given file name.
    :rtype: str
    """
    return os.path.join(TEST_DIR, file_name)



"""Logging utilities and constants.
"""
# Built-in
import pickle
import logging
import logging.handlers
import struct
import threading
import os
import sys
import tempfile
import errno
import time
# Python 2/3 compatiblity
if sys.version_info[0] == 2:
    import SocketServer
else:
    import socketserver as SocketServer

# Internal
from .constants import USER_DIR

VERBOSE_ENV_VAR = 'NXT_VERBOSE'

"""Logging levels are as follows, with custom levels marked with *
CRITICAL    - 50
ERROR       - 40
*GRAPHERROR - 39
WARNING     - 30
INFO        - 20
*NODEOUT    - 19
*EXECINFO   - 18
*COMPINFO   - 17
*SOCKET     - 16
DEBUG       - 10
NOTSET      - 0
"""
SOCKET = 16
GRAPHERROR = 39
NODEOUT = 19
EXECINFO = 18
COMPINFO = 17

COLORS = {
    SOCKET: 'pink',
    GRAPHERROR: 'red',
    NODEOUT: 'light gray',
    EXECINFO: '#039be5',  # Pale blue
    COMPINFO: 'light blue',
    logging.DEBUG: 'white',
    logging.INFO: 'green',
    logging.WARNING: 'orange',
    logging.ERROR: 'red',
    logging.CRITICAL: 'purple',
}


# logging setup
# NOTE probably neeed to be custom logger objects rather than our custom levels
class NXTLogger(logging.getLoggerClass()):
    """Implements custom logging levels for nxt.
    """
    def __init__(self, name):
        super(NXTLogger, self).__init__(name)
        logging.addLevelName(SOCKET, "SOCKET")
        logging.addLevelName(EXECINFO, "EXECUTE")
        logging.addLevelName(GRAPHERROR, "GRAPH ERROR")
        logging.addLevelName(COMPINFO, "COMPOSITE")
        logging.addLevelName(NODEOUT, "NODEOUT")

    def socket(self, message, *args, **kwargs):
        if self.isEnabledFor(SOCKET):
            self._log(SOCKET, message, args, **kwargs)

    def execinfo(self, message, *args, **kwargs):
        if self.isEnabledFor(EXECINFO):
            self._log(EXECINFO, message, args, **kwargs)

    def grapherror(self, message, *args, **kwargs):
        if self.isEnabledFor(GRAPHERROR):
            self._log(GRAPHERROR, message, args, **kwargs)

    def compinfo(self, message, *args, **kwargs):
        if self.isEnabledFor(COMPINFO):
            self._log(COMPINFO, message, args, **kwargs)

    def nodeout(self, message, *args, **kwargs):
        if self.isEnabledFor(NODEOUT):
            self._log(NODEOUT, message, args, **kwargs)

    def _log(self, level, message, args, **kwargs):
        if 'links' in list(kwargs.keys()):
            link_paths = kwargs.pop('links')
            kwargs['extra'] = {'links': link_paths}
        super(NXTLogger, self)._log(level, message, args, **kwargs)


nxt_formatter = logging.Formatter("%(levelname)s: %(message)s")


def track_log_file(path):
    """
    Track given log at `path`, via meta log.
    Clean up old log files tracked by meta log.
    """
    logger = logging.getLogger(__name__)
    meta_log_location = os.path.join(USER_DIR, '.nxt_meta_log')
    remove_age = 60*60*48  # in seconds (currently 48 hours)

    existing_logs = []
    try:
        with open(meta_log_location, 'r') as meta_log_file:
            # stripping of trailing newlines
            existing_logs = [line.rstrip() for line in meta_log_file.readlines()]
    except IOError:
        logger.exception('Error tracking log file!')
        pass
    # Remove old logs
    for log_path in existing_logs[:]:
        is_file = os.path.isfile(log_path)
        age = 0.0
        try:
            mod_time = os.path.getmtime(log_path)
            age = time.time() - mod_time
        except Exception as err:
            try:
                if err.errno is errno.ENOENT:
                    # This is expected behavior. Because files log to tempdir,
                    # they can be removed at any time by the OS or user.
                    age = float("inf")
            except AttributeError:
                pass
        if is_file and age > remove_age:
            try:
                os.remove(log_path)
            except Exception as err:
                logger.exception(err)

        elif not is_file or age > remove_age:
            existing_logs.remove(log_path)
            logger.info('Removing stale log "{}"'.format(log_path))
    # add given log and write meta log
    existing_logs.append(path)
    try:
        with open(meta_log_location, 'w') as meta_log_file:
            # add trailing newlines to make meta log human readable.
            meta_log_file.writelines([line + '\n' for line in existing_logs])
    except IOError:
        logger.exception('Failed to track "{}" log in meta_log'.format(path))


def initial_setup():
    """
    module-wide logging boilerplate.
    Sets logging level.
    Builds and sets up default handlers.
    """
    if logging.getLoggerClass() is NXTLogger:
        logging.info('nxt logger already setup!')
        return
    logging.setLoggerClass(NXTLogger)
    root_logger = logging.getLogger('nxt')
    root_logger.setLevel(logging.DEBUG)
    root_logger.propagate = False
    null_handler = logging.NullHandler()
    root_logger.addHandler(null_handler)

    env_verbosity = os.environ.get(VERBOSE_ENV_VAR)
    global verbose_handler
    verbose_handler = None
    if env_verbosity is None:
        return
    verbose_handler = set_verbosity(env_verbosity)


def set_verbosity(level):
    """Set the verbosity of nxt logging. Controls what to output to stderr.
    3 legal levels specified by integer. NOTE node prints are not affected.\n
    0 - Nothing\n
    1 - EXECINFO and higher\n
    2 - Everything\n

    :param level: level to set stderr verbosity to: 0, 1, or 2
    :type level: int
    :return: stderr handler, if there is one
    :rtype: logging.StreamHandler or None
    """
    # TODO when the swap to dedicated loggers is made. Level 1 should install
    # a stderr handler only to the node execution logger.
    root_logger = logging.getLogger('nxt')
    global verbose_handler
    if level == 'socket':
        _log_port = logging.handlers.DEFAULT_TCP_LOGGING_PORT
        socket_handler = logging.handlers.SocketHandler('localhost', _log_port)
        # Add socket handler
        root_logger.addHandler(socket_handler)
        return
    level = int(level)
    if level == 0:
        if not verbose_handler:
            return
        root_logger.removeHandler(verbose_handler)
        return
    if level > 0:
        if verbose_handler:
            stderr_handler = verbose_handler
        else:
            stderr_handler = logging.StreamHandler(sys.stderr)
            stderr_handler.setFormatter(nxt_formatter)
            root_logger.addHandler(stderr_handler)
        if level == 1:
            stderr_handler.setLevel(EXECINFO)
        return stderr_handler


# Dictionary mapping filenames to filehandlers, allows them to be removed.
SESSION_LOG_HANDLERS = {}


def make_session_log(filename=None):
    """Begin saving log at `filename` if given, otherwise
    track to a generated file inside default temporary directory.

    :param filename: file to log to, defaults to None
    :type filename: str, optional
    :return: file being logged to, `filename` if given.
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    if not filename:
        filename = get_new_session_log_filename()
    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(nxt_formatter)
    root_logger = logging.getLogger('nxt')
    root_logger.addHandler(file_handler)
    SESSION_LOG_HANDLERS[filename] = file_handler
    logger.info("logging to {}".format(filename))
    os.chmod(filename, 0o777)
    track_log_file(filename)
    return filename


def get_new_session_log_filename():
    """Get a new log filename for an nxt session inside the default logs directory.
    Attempts to create logs directory.

    :return: filename to use for session log.
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    log_dir = os.path.join(tempfile.gettempdir(), 'nxt_logs')
    try:
        os.makedirs(log_dir)
        os.chmod(log_dir, 0o777)
    except OSError as err:
        if err.errno is errno.EEXIST:
            logger.debug("Not creating logs directory because it exists.")
        else:
            raise
    log_filename_template = "nxt_session_{time}.log"
    log_filename = log_filename_template.format(time=str(int(time.time())))
    return os.path.join(log_dir, log_filename)


def stop_session_log(filename):
    """Stops outputting log to given filename.

    :param filename: filename to stop outputting to.
    :type filename: str, required
    """
    handler = SESSION_LOG_HANDLERS.get(filename)
    if not handler:
        return
    root_logger = logging.getLogger('nxt')
    root_logger.removeHandler(handler)


def make_link(value):
    """Wrap given `value` in html href syntax.

    :param value: value to wrap
    :type value: str
    """
    # TODO: Maybe someone smarter can find the answer but this thread says
    #  you can not set link colors in the qss
    #  https://stackoverflow.com/questions/13416183/qt-stylesheet-how-to-set-color-of-qpalettelink-and-qplattelinkvisited
    return '<a style="color: #039be5" href=\"{}\">{}</a>'.format(value, value)


class LogRecordStreamHandler(SocketServer.StreamRequestHandler):
    """Lifted from: https://docs.python.org/2.7/howto/logging-cookbook.html
    Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.un_pickle(chunk)
            level_name = obj.get('levelname')
            if level_name:
                obj['levelname'] = 'remote.' + obj['levelname']
            record = logging.makeLogRecord(obj)
            self.handle_log_record(record)

    def un_pickle(self, data):
        return pickle.loads(data)

    def handle_log_record(self, record):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name
        logger = logging.getLogger(name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)


class LogRecordSocketReceiver(SocketServer.ThreadingTCPServer):
    """Lifted from: https://docs.python.org/2.7/howto/logging-cookbook.html
    Simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = 1

    def __init__(self, host='localhost',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                 handler=LogRecordStreamHandler):
        SocketServer.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None
        self.kill = False

    def serve_until_stopped(self):
        import select
        abort = 0
        while not abort and not self.kill:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [],
                                       self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort


record_receiver = None


def startup_log_socket(stream_handler=LogRecordStreamHandler):
    """Starts threaded logging socket receiver, if using a visual app a
    thread safe stream handler should be provided.

    :param stream_handler: Logging stream handler default LogRecordStreamHandler
    :return: LogRecordSocketReceiver()
    """
    global record_receiver
    stream_handler = stream_handler or LogRecordStreamHandler
    record_receiver = LogRecordSocketReceiver(handler=stream_handler)
    logging.info('About to start logging TCP server...')
    tcp_thread = threading.Thread(target=record_receiver.serve_until_stopped)
    tcp_thread.start()


def shutdown_log_socket():
    global record_receiver
    if record_receiver:
        record_receiver.kill = True



# Built-in
import unittest
import os

# Internal
from nxt import stage, DATA_STATE, tokens
from nxt.session import Session


class TestPathTokens(unittest.TestCase):
    def test_no_change_full_path(self):
        """Give a full path and verify that it is not mangled by resolution.
        """
        test_attr = 'no_change'
        in_val = '${path::' + __file__ + '}'
        expected = __file__.replace(os.path.sep, '/')
        test_stage = stage.Stage()
        test_node, _ = test_stage.add_node(layer=0)
        test_node = test_node[0]
        test_stage.add_node_attr(test_node, test_attr, {'value': in_val},
                                 test_stage.top_layer)
        result = test_stage.get_node_attr_value(test_node, test_attr,
                                                test_stage.top_layer)
        self.assertEqual(expected, result)

    def test_sibling_path(self):
        """Test that resolution of a filename only returns a file next to it.

        NOTE: this test depends on the test folder structure. It is using a
        test graph `TokenSiblingTest.nxt` that must be a direct sibling
        to this test file. The contents of that file are also depended on.
        If you find one of these files without the other, something is wrong.
        """
        base_path = __file__.replace(os.path.sep, '/')
        if base_path.endswith('.pyc'):
            base_path = base_path[:-1]
        base_name = os.path.basename(base_path)
        in_val = '${file::' + base_name + '}'
        expected = base_path
        test_attr = 'my_file'
        my_dir = os.path.dirname(base_path)
        os.chdir(my_dir)
        test_stage = Session().load_file(filepath='TokenSiblingTest.nxt')
        comp_layer = test_stage.build_stage()
        test_node = test_stage.top_layer.lookup('/tester')
        test_stage.add_node_attr(test_node, test_attr, {'value': in_val},
                                 test_stage.top_layer)
        result = test_stage.get_node_attr_value(test_node, test_attr,
                                                comp_layer)
        self.assertEqual(expected, result)

    def test_file_token(self):
        base_path = __file__.replace(os.path.sep, '/')
        if base_path.endswith('.pyc'):
            base_path = base_path[:-1]
        my_dir = os.path.dirname(base_path)
        os.chdir(my_dir)
        test_stage = Session().load_file(filepath='FileTokenSub.nxt')
        comp_layer = test_stage.build_stage()
        fake_path = os.path.join(my_dir, 'notafile.txt')
        fake_path = fake_path.replace(os.path.sep, '/')
        real_path = os.path.join(my_dir, 'real_file.txt')
        real_path = real_path.replace(os.path.sep, '/')
        test_node = test_stage.top_layer.lookup('/tokens')
        file_fake = test_stage.get_node_attr_value(test_node, 'file_fake',
                                                   comp_layer,
                                                   resolved=True)
        self.assertEqual('', file_fake)
        file_real = test_stage.get_node_attr_value(test_node, 'file_real',
                                                   comp_layer,
                                                   resolved=True)
        self.assertEqual(real_path, file_real)
        path_fake = test_stage.get_node_attr_value(test_node, 'path_fake',
                                                   comp_layer,
                                                   resolved=True)
        self.assertEqual(fake_path, path_fake)
        path_real = test_stage.get_node_attr_value(test_node, 'path_real',
                                                   test_stage.top_layer,
                                                   resolved=True)
        self.assertEqual(real_path, path_real)

    def test_contents_token(self):
        base_path = __file__.replace(os.path.sep, '/')
        if base_path.endswith('.pyc'):
            base_path = base_path[:-1]
        my_dir = os.path.dirname(base_path)
        os.chdir(my_dir)
        test_stage = Session().load_file(filepath='FileTokenSub.nxt')
        with open('real_file.txt', 'r') as fp:
            expected_code = fp.read()
        top = test_stage.top_layer
        node = top.lookup('/tokens')
        resolved = DATA_STATE.RESOLVED
        found_code = test_stage.get_node_code_string(node=node,
                                                           layer=top,
                                                           data_state=resolved)
        self.assertEqual(expected_code, found_code)


class TestAttrRefTokens(unittest.TestCase):
    def test_attr_ref_building(self):
        test_in_exp = {
            'stuff.thing': '${stuff.thing}',
            'all/other/stuff.another': '${all/other/stuff.another}'
        }
        for inp, exp in test_in_exp.items():
            result = tokens.make_token_str(inp)
            self.assertEqual(exp, result)



# Builtin
import os
from collections import OrderedDict
import json
import logging
import shutil
import pickle

# Internal
from . import clean_json
from .constants import GRAPH_VERSION

#
#  By design you must not import code from nxt, the code here must be a
#  historical record of the code. This file is very not "dry", it is a
#  script with basically zero dependencies.
#
logger = logging.getLogger(__name__)

PRE_RELEASE = (0, 45)


class FileConverter(object):
    def __init__(self, file_data):
        """Determines the version number of the given file. And steps
        through our legacy converters until the version number matches the
        current NXT version. We ignore the major version number and the bug
        fix number.
        :param file_data: Dict of file data from NxtIo
        """
        self.file_data = file_data
        self.version = self.determine_version()
        nxt_version = int(GRAPH_VERSION.VERSION_TUPLE[1])
        if self.version == PRE_RELEASE:
            self.file_data = FILE_VERSIONS[PRE_RELEASE].get_converted(self.file_data)
            self.version = self.determine_version()
        v = self.version
        while True:
            if FILE_VERSIONS.get(v):
                converter = FILE_VERSIONS[v]
                self.file_data = converter.get_converted(self.file_data)
            if v == nxt_version:
                break
            # Step version number by 1
            v += 1

    def determine_version(self):
        """Strips off the bug fix number then returns the major version
        number and the feature number.
        :return: Version number #.##
        """
        file_version = self.file_data.get('version', '0.45.0')
        major_version, minor_version = [int(n) for n in
                                        file_version.split('.')[0:2]]
        if major_version == 0 and minor_version >= 39:
            # v0.39.0 was the last converter for pre-release v0.xx
            return PRE_RELEASE
        elif major_version == 0 and minor_version < 39:
            raise IOError('File version v{} is too old  for conversion. '
                          'Please open in nxt v0.45 and save again in order '
                          'to open it in '
                          '{}'.format(file_version,
                                      GRAPH_VERSION.VERSION_STR))
        return minor_version

    @classmethod
    def get_converted_data(cls, file_data):
        """Sets up converter object and returns the converted data. Note the
        input data and output data may be the same if no conversion is
        deemed necessary.
        :param file_data: Dict of file data from NxtIo
        :return: Updated file data dict
        """
        save_converter = cls(OrderedDict(file_data))
        return save_converter.file_data


class LegacyFileConverter(object):
    def __init__(self, file_data, from_version, to_version):
        """Base class used by all converters
        :param file_data: Dict of file data from NxtIo
        :param from_version: String of input version, used for debugging
        :param to_version: String of output version, used for debugging."""
        self.file_data = file_data
        self.converted = False
        self.from_version = from_version
        self.to_version = to_version

    def convert(self):
        """Meant to be overloaded with conversion logic."""
        logger.info("Converting graph {} --> {}".format(self.from_version,
                                                        self.to_version))

    @classmethod
    def get_converted(cls, file_data):
        """Kicks off the conversion of the data and returns the converted data.
        :param file_data: Dict of file data from NxtIo
        :return: Updated file data dict
        """
        converter = cls(file_data)
        converter.convert()
        return converter.file_data


class V1(LegacyFileConverter):
    def __init__(self, file_data):
        """Converts from v0.38.x to 'v1.0.0'
        :param file_data: Dict of save data.
        """
        super(V1, self).__init__(file_data, 'v0.45.x', 'v1.0.0')

    def convert(self):
        super(V1, self).convert()
        keys = list(self.file_data.keys())
        if 'children' in keys:
            children = self.file_data.pop('children')
        else:
            children = []
        for node_data in children:
            parent_path = '/'
            self.unpack_node_data(node_data, parent_path)
        self.file_data['version'] = '1.0.0'
        if 'data' in keys:
            self.file_data['meta_data'] = self.file_data.pop('data')
        self.arrange_node_data()
        self.drop_nodes()
        self.deep_order_data(self.file_data)

    def unpack_node_data(self, node_data, parent_path):
        flat_data = OrderedDict()
        widget = node_data['data'].get('widget_type')
        if self.has_opinion(widget):
            flat_data['widget_type'] = widget
        exec_in = node_data.get('execute_in')
        if self.has_opinion(exec_in):
            flat_data['execute_in'] = exec_in
        inst = node_data['data'].get('instance')
        if self.has_opinion(inst):
            flat_data['instance'] = inst
        enabled = node_data['data'].get('enabled')
        if enabled is not None:
            flat_data['enabled'] = enabled
        comp = node_data['data'].get('compute')
        if self.has_opinion(comp):
            flat_data['code'] = comp
        comm = node_data.get('comment')
        if self.has_opinion(comm):
            flat_data['comment'] = comm
        attrs = node_data['data'].get('attributes')
        if self.has_opinion(attrs):
            flat_data['attrs'] = attrs
        child_order = node_data.get('execute_order')
        if self.has_opinion(child_order):
            flat_data['child_order'] = child_order
        path = parent_path + node_data['name']
        if path in self.file_data['data'].get('start_nodes', []):
            flat_data['start_point'] = True
        # Make sure we have a node's dict
        self.file_data.setdefault('nodes', OrderedDict())
        self.file_data['nodes'][path] = flat_data
        for child_data in node_data.get("children", []):
            self.unpack_node_data(child_data, path+'/')

    def arrange_node_data(self):
        try:
            unsorted_paths = list(self.file_data['nodes'].keys())
        except KeyError:
            return
        paths_by_depth = {}
        max_depth = 0
        for path in unsorted_paths:
            depth = path.count('/')
            paths_by_depth.setdefault(depth, [])
            paths_by_depth[depth].append(path)
            if depth > max_depth:
                max_depth = depth
        for depth in list(paths_by_depth.keys()):
            paths_by_depth[depth].sort()

        sorted_paths = []

        def handle_path_at_depth(cur_path, cur_depth):
            for d in range(cur_depth + 1, max_depth + 1):
                for lower_path in paths_by_depth.get(d, []):
                    if self.is_ancestor(lower_path, cur_path):
                        if lower_path not in sorted_paths:
                            sorted_paths.append(lower_path)
                            handle_path_at_depth(lower_path, d)

        for depth in range(1, max_depth + 1):
            for path in paths_by_depth.get(depth, []):
                if path not in sorted_paths:
                    sorted_paths.append(path)
                    handle_path_at_depth(path, depth)

        # sorted_paths = sorted(unsorted_paths)
        new_nodes = OrderedDict()
        for path in sorted_paths:
            new_nodes[path] = self.file_data['nodes'][path]
        self.file_data['nodes'] = new_nodes

    def deep_order_data(self, node_dict):
        """Recursively orders the ordered dicts"""
        # order keys
        keys_order = ['version', 'alias', 'name', 'comment', 'color', 'mute',
                      'solo', 'references', 'path', 'instance',
                      'enabled', 'widget_type', 'meta_data', 'execute_in',
                      'execute_order', 'nodes', 'attributes', 'code',
                      'real_path']
        for key in keys_order:
            if key in list(node_dict.keys()):
                node_dict[key] = node_dict.pop(key)
        # iterate over deeper values
        for value in node_dict.values():
            if isinstance(value, dict):
                self.deep_order_data(value)
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self.deep_order_data(item)

    def drop_nodes(self):
        remove = []
        for path, node in self.file_data.get('nodes', {}).items():
            name = self.get_name_from_path(path)
            parent_path = self.get_parent_path(path)
            parent = self.file_data['nodes'].get(parent_path)
            if parent and name not in parent.get('child_order', []):
                remove += [path]
        for path in remove:
            logger.info('Dropping: ' + path)
            self.file_data['nodes'].pop(path)

    @staticmethod
    def has_opinion(attr_value):
        return attr_value not in (None, [], {})

    def is_ancestor(self, path, ancestor_path):
        path = self._add_path_terminator(path)
        ancestor_path = self._add_path_terminator(ancestor_path)
        return path.startswith(ancestor_path)

    @staticmethod
    def _add_path_terminator(path):
        if not path.endswith('/'):
            path += '/'
        return path

    @staticmethod
    def get_parent_path(path):
        parent_path = '/'.join(path.split('/')[:-1])
        return parent_path

    @staticmethod
    def get_name_from_path(path):
        name = path.split('/')[-1]
        return name


class V109(LegacyFileConverter):
    def __init__(self, file_data):
        """Converts from v1.x.x to 'v1.10.0'
        :param file_data: Dict of save data.
        """
        super(V109, self).__init__(file_data, 'v0.9.x', 'v1.10.0')

    def convert(self):
        super(V109, self).convert()
        self.update_refs()
        self.update_meta_data()
        self.arrange_node_data()
        self.deep_order_data(self.file_data)

    def update_refs(self):
        ref_dicts = self.file_data.get('references')
        if not ref_dicts:
            return
        refs = []
        self.file_data.pop('references')
        meta_data = self.file_data.get('meta_data', {})
        comp_overrides = self.file_data.get('comp_overrides', {})
        aliases = meta_data.get('aliases', {})
        colors = meta_data.get('colors', {})
        for ref in ref_dicts:
            path = ref.get('path')
            if not path:
                continue
            r_dat = comp_overrides.get(path, {})
            mute = ref.get('mute', None)
            solo = ref.get('solo', None)
            alias = ref.get('alias', None)
            color = ref.get('color', None)
            if mute is not None:
                r_dat['mute'] = mute
            if solo is not None:
                r_dat['solo'] = solo
            if alias is not None:
                aliases[path] = alias
            if color is not None:
                colors[path] = color
            refs.append(path)
            if r_dat:
                comp_overrides[path] = r_dat
        if refs:
            self.file_data['references'] = refs
        else:
            return
        if comp_overrides:
            self.file_data['comp_overrides'] = comp_overrides
        if aliases:
            meta_data['aliases'] = aliases
        if colors:
            meta_data['colors'] = colors
        if meta_data:
            self.file_data['meta_data'] = meta_data

    def update_meta_data(self):
        meta_data = self.file_data.get('meta_data')
        if not meta_data:
            return
        if meta_data.get('position_data'):
            positions = meta_data.pop('position_data')
            meta_data['positions'] = positions
        if meta_data.get('collapse_data'):
            collapse = meta_data.pop('collapse_data')
            meta_data['collapse'] = collapse

    def arrange_node_data(self):
        try:
            unsorted_paths = list(self.file_data['nodes'].keys())
        except KeyError:
            return
        paths_by_depth = {}
        max_depth = 0
        for path in unsorted_paths:
            depth = path.count('/')
            paths_by_depth.setdefault(depth, [])
            paths_by_depth[depth].append(path)
            if depth > max_depth:
                max_depth = depth
        for depth in list(paths_by_depth.keys()):
            paths_by_depth[depth].sort()

        sorted_paths = []

        def handle_path_at_depth(cur_path, cur_depth):
            for d in range(cur_depth + 1, max_depth + 1):
                for lower_path in paths_by_depth.get(d, []):
                    if self.is_ancestor(lower_path, cur_path):
                        if lower_path not in sorted_paths:
                            sorted_paths.append(lower_path)
                            handle_path_at_depth(lower_path, d)

        for depth in range(1, max_depth + 1):
            for path in paths_by_depth.get(depth, []):
                if path not in sorted_paths:
                    sorted_paths.append(path)
                    handle_path_at_depth(path, depth)

        # sorted_paths = sorted(unsorted_paths)
        new_nodes = OrderedDict()
        for path in sorted_paths:
            new_nodes[path] = self.file_data['nodes'][path]
        self.file_data['nodes'] = new_nodes

    def deep_order_data(self, node_dict):
        """Recursively orders the ordered dicts"""
        # order keys
        keys_order = ['version', 'alias', 'name', 'comment', 'color', 'mute',
                      'solo', 'references', 'comp_overrides', 'path',
                      'instance', 'enabled', 'widget_type', 'meta_data',
                      'aliases', 'colors', 'collapse', 'positions',
                      'execute_in', 'execute_order', 'nodes', 'attributes',
                      'code', 'real_path']
        for key in keys_order:
            if key in list(node_dict.keys()):
                node_dict[key] = node_dict.pop(key)
        # iterate over deeper values
        for value in node_dict.values():
            if isinstance(value, dict):
                self.deep_order_data(value)
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self.deep_order_data(item)

    def is_ancestor(self, path, ancestor_path):
        path = self._add_path_terminator(path)
        ancestor_path = self._add_path_terminator(ancestor_path)
        return path.startswith(ancestor_path)

    @staticmethod
    def _add_path_terminator(path):
        if not path.endswith('/'):
            path += '/'
        return path


class V114(LegacyFileConverter):
    def __init__(self, file_data):
        """Converts from v1.x.x to 'v1.15.0'
        :param file_data: Dict of save data.
        """
        super(V114, self).__init__(file_data, 'v0.14.x', 'v1.15.0')

    def convert(self):
        super(V114, self).convert()
        self.file_data['version'] = '1.15'
        if self.file_data.get('mute') is False:
            self.file_data.pop('mute')
        if self.file_data.get('solo') is False:
            self.file_data.pop('solo')
        if self.file_data.get('references') is []:
            self.file_data.pop('references')
        self.arrange_node_data()
        self.deep_order_data(self.file_data)

    def arrange_node_data(self):
        try:
            unsorted_paths = list(self.file_data['nodes'].keys())
        except KeyError:
            return
        paths_by_depth = {}
        max_depth = 0
        for path in unsorted_paths:
            depth = path.count('/')
            paths_by_depth.setdefault(depth, [])
            paths_by_depth[depth].append(path)
            if depth > max_depth:
                max_depth = depth
        for depth in list(paths_by_depth.keys()):
            paths_by_depth[depth].sort()

        sorted_paths = []

        def handle_path_at_depth(cur_path, cur_depth):
            for d in range(cur_depth + 1, max_depth + 1):
                for lower_path in paths_by_depth.get(d, []):
                    if self.is_ancestor(lower_path, cur_path):
                        if lower_path not in sorted_paths:
                            sorted_paths.append(lower_path)
                            handle_path_at_depth(lower_path, d)

        for depth in range(1, max_depth + 1):
            for path in paths_by_depth.get(depth, []):
                if path not in sorted_paths:
                    sorted_paths.append(path)
                    handle_path_at_depth(path, depth)

        # sorted_paths = sorted(unsorted_paths)
        new_nodes = OrderedDict()
        for path in sorted_paths:
            new_nodes[path] = self.file_data['nodes'][path]
        self.file_data['nodes'] = new_nodes

    def deep_order_data(self, node_dict):
        """Recursively orders the ordered dicts"""
        # order keys
        keys_order = ['version', 'alias', 'name', 'comment', 'color', 'mute',
                      'solo', 'references', 'comp_overrides', 'path',
                      'instance', 'enabled', 'widget_type', 'meta_data',
                      'aliases', 'colors', 'collapse', 'positions',
                      'execute_in', 'execute_order', 'nodes', 'attributes',
                      'code', 'real_path']
        for key in keys_order:
            if key in list(node_dict.keys()):
                node_dict[key] = node_dict.pop(key)
        # iterate over deeper values
        for value in node_dict.values():
            if isinstance(value, dict):
                self.deep_order_data(value)
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self.deep_order_data(item)

    def is_ancestor(self, path, ancestor_path):
        path = self._add_path_terminator(path)
        ancestor_path = self._add_path_terminator(ancestor_path)
        return path.startswith(ancestor_path)

    @staticmethod
    def _add_path_terminator(path):
        if not path.endswith('/'):
            path += '/'
        return path


class V115(LegacyFileConverter):
    def __init__(self, file_data):
        """Converts from 1.15 to 1.16
        :param file_data: Dict of save data.
        """
        super(V115, self).__init__(file_data, '0.15', '1.16')

    def convert(self):
        super(V115, self).convert()
        self.file_data['version'] = '1.16'
        self.file_data = self.order_save_data(self.file_data)

    def order_save_data(self, file_dict):
        keys_order = ('version', 'alias', 'color', 'mute', 'solo',
                      'references', 'comp_overrides', 'meta_data', 'nodes',
                      'real_path')
        result = OrderedDict()
        data_keys = list(file_dict.keys())
        for key in keys_order:
            if key not in data_keys:
                continue
            data_keys.remove(key)
            if key == 'nodes':
                result['nodes'] = self.order_nodes_dict(file_dict['nodes'])
            elif key == 'meta_data':
                meta_data = file_dict['meta_data']
                result['meta_data'] = self.order_meta_data(meta_data)
            elif key == 'comp_overrides':
                overs = file_dict['comp_overrides']
                ordered_overs = OrderedDict(sorted(overs.items(),
                                                   key=lambda x: x[0]))
                result['comp_overrides'] = ordered_overs
            else:
                result[key] = file_dict[key]
        # leftovers
        for key in data_keys:
            result[key] = file_dict[key]
        return result

    def order_nodes_dict(self, nodes_dict):
        unsorted = list(nodes_dict.keys())
        out_nodes = OrderedDict()
        sorted_paths = sorted(unsorted)
        for path in sorted_paths:
            out_nodes[path] = self.order_node_data(nodes_dict[path])
        return out_nodes

    def order_node_data(self, node_data):
        out_node = OrderedDict()
        keys_order = ('start_point', 'instance', 'execute_in', 'child_order',
                      'enabled', 'comment', 'attrs', 'code')
        for key in keys_order:
            if key not in node_data:
                continue
            if key == 'attrs':
                old_attrs = node_data['attrs']
                attrs_order = sorted(list(old_attrs.keys()))
                result_attrs = OrderedDict()
                for attr_name in attrs_order:
                    old_sub_attrs = old_attrs[attr_name]
                    sub_attrs = OrderedDict(sorted(old_sub_attrs.items(),
                                                   key=lambda x: x[0]))
                    result_attrs[attr_name] = sub_attrs
                out_node['attrs'] = result_attrs
            else:
                out_node[key] = node_data[key]
        return out_node

    def order_meta_data(self, meta_data):
        keys_order = ('aliases', 'colors', 'positions', 'collapse')
        result = OrderedDict()
        data_keys = list(meta_data.keys())
        for key in keys_order:
            if key not in data_keys:
                continue
            data_keys.remove(key)
            result[key] = OrderedDict(sorted(meta_data[key].items(),
                                             key=lambda x: x[0]))
        # leftovers
        for key in data_keys:
            result[key] = meta_data[key]
        return result


FILE_VERSIONS = {
    PRE_RELEASE: V1,
    9: V109,
    14: V114,
    15: V115
}


def cli_file_convert(file_path, replace=False):
    """Simple logic for converting files from the CLI"""
    file_path = os.path.realpath(os.path.expanduser(file_path))
    logger.info("Trying to convert: " + file_path)
    with open(file_path, 'r') as f:
        json_data = json.load(f, object_hook=clean_json._byteify)
        old_file_data = clean_json.load(json_data)
    file_data = FileConverter.get_converted_data(old_file_data)
    out_file_name = os.path.basename(file_path)
    out_file_name, _, ext = out_file_name.rpartition('.')
    if replace:
        ending = "."
    else:
        ending = "_CONVERTED."
    out_file_name = out_file_name + ending + ext
    out_file_dir = os.path.dirname(file_path)
    out_file_path = os.path.join(out_file_dir, out_file_name)
    with open(out_file_path, 'w') as out_file:
        json.dump(file_data, out_file, indent=4, sort_keys=False,
                  separators=(',', ': '))
        logger.info('Successfully saved "' + out_file_path + '"')



# Builtin
import unittest
import os
import json
import subprocess
import sys

# Internal
import nxt
from nxt.session import Session
from nxt.test import get_test_file_path


class CLI(unittest.TestCase):
    '''The test graph writes /cli_test.my_attr to the location specified by
    The env var $_TEST_WRITE_PATH, allowing verification of paramter overrides
    by reading the resulting file.'''
    test_graph = get_test_file_path('cli_test.nxt')

    def setUp(self):
        self.test_path = get_test_file_path('delete_me')
        # Test graph writes a file to the location of this env var.
        self.test_env = os.environ.copy()
        self.test_env.update({'_TEST_WRITE_PATH': self.test_path})

    def tearDown(self):
        try:
            os.remove(self.test_path)
        except FileNotFoundError:
            pass

    def read_test_path(self):
        with open(self.test_path, 'r') as fp:
            return fp.read()

    def test_basic_cli_exec(self):
        """Basic call proves test graph works as expected"""
        ret = subprocess.call([sys.executable, '-m', 'nxt.cli', 'exec',
                               self.test_graph], env=self.test_env)
        self.assertEqual(0, ret)
        self.assertTrue(self.read_test_path() == 'default')

    def test_cli_parameters(self):
        """Overwriting attributes from the cli"""
        subprocess.call([sys.executable, '-m', 'nxt.cli', 'exec',
                        self.test_graph, '-p', '/cli_test.my_attr',
                        'from_cli'], env=self.test_env)
        self.assertTrue(self.read_test_path() == 'from_cli')

    def test_file_parameters(self):
        """Overwriting attributes from parameters file"""
        tmp_params_path = get_test_file_path('test_params')
        with open(tmp_params_path, "w+") as fp:
            json.dump({'/cli_test.my_attr': 'from_file'}, fp)
        subprocess.call([sys.executable, '-m', 'nxt.cli', 'exec',
                        self.test_graph, '-p', tmp_params_path],
                        env=self.test_env)
        os.remove(tmp_params_path)
        self.assertTrue(self.read_test_path() == 'from_file')


class PythonEntry(unittest.TestCase):

    graph_path = get_test_file_path("StageRuntimeScope.nxt")

    def test_simple_python_entry(self):
        nxt.execute_graph(self.graph_path)

    def test_python_entry(self):
        my_session = Session()
        rtl = my_session.execute_graph(self.graph_path)
        self.assertIsNotNone(rtl)
